/* (c) Peter Horsman, 2002
 * Pete (dooby@bits.bris.ac.uk / http://bits.bris.ac.uk/dooby/)
 */

#include "gba.h"
#include "regs.h"
#include "interrupt.h"

#include "mode5lib.h"

#define CIRCLE5(x, y, r) ellipse5(x, y, ((r) >> 1) + ((r) >> 4), r)

MULTIBOOT

int AgbMain(void) {
  int x, y, i;
  u16 col;

  // Set GBA mode 4 with sprites enabled.
  DISPCNT = 0x0405;	// NB don't enable window flag unless it's set up :)

  // Initialise interrupts to allow A+B+Start+Select to hard reset.
  interrupt_init();
  KEYCNT = KEY_AND | KEY_IRQ | PAD_A | PAD_B | PAD_START | PAD_SELECT;
  IE |= INT_KEY;
  IME = 1;

  //gba_clsUnroll2(0);

  // UNCOMMENT BELOW TO WATCH THE ROTATION TAKING PLACE.
  
  // Fill screen w/ horizontal stripes.
  /*for(y = 0; y < 128; y ++) {
    col = gba_rand() & 0x7fff;
    for(x = 0; x < 160; x ++) {
      ((unsigned short *)0x06000000)[y * 160 + x] = col;
    }
    gba_vsync();
  }*/

  // Fill screen with random noise.
  //for(x = 0; x < 20 * 1024; x ++) ((unsigned short *)0x06000000)[x] = gba_rand() & 0x7fff;

  // Horizontal scale (2 << 8 means each row is half width).
  //BG2PA = 1 << 8;	// default 1 << 8
  // Horizontal shear (1 << 8 means move 1 pixel left each pixel down).
  //BG2PB = 0 << 8;	// default 0 << 8
  // Vertical shear (1 << 8 means move 1 pixel up every pixel right).
  //BG2PC = 0 << 8;	// default 0 << 8
  // Y scale (2 << 8 means each line is half height).
  //BG2PD = 1 << 8;	// default 1 << 8

  // Rotate/scale to make horizontal lines look vertical.
  /*for(x = 0; x >= -90; x --) {
    BG2PA = gba_cos(x) << 1;
    BG2PB = gba_sin(x) << 1;
    BG2PC = -gba_sin(x) << 1;
    BG2PD = gba_cos(x) << 1;
    gba_vsync();
  }*/

  // Shift screen 160 pixels down
  /*for(x = 0; x < 160; x ++) {
    BG2X = x * 256;
    gba_vsync();
  }*/

  // Decrementing BG2PA now has effect of increasing downward vertical shear as
  // you go right.
  // Increasing BG2PB now seems to decrease Y scale.
  // Increasing BG2PC now seems to increase X scale.

  // Scale screen right to fit. (128 / 240 * 256 = 136.5333)
  /*for(x = 256; x >= 136; x --) {
    BG2PC = x;
    gba_vsync();
  }*/

  // Have some fun.
  BG2PA = 0;
  BG2PB = -1 << 8;
  BG2PC = 137;
  BG2PD = 0;
  BG2X = 159 << 8;

  clear5(0x3c003c00);

  // Yellow box.
  for(x = 0; x < 128; x ++) {
    put_pixel5(x, 0, 0x03ff);
    put_pixel5(x, 159, 0x03ff);
  }
  for(y = 0; y < 160; y ++) {
    put_pixel5(0, y, 0x03ff);
    put_pixel5(127, y, 0x03ff);
  }


  for(x = 0; x < 128; x ++) {
    put_pixel5(x, 80, 0x001f);
  }
  for(y = 0; y < 160; y ++) {
    put_pixel5(64, y, 0x03e0);
  }
  for(i = 0; i < 128; i ++) {
    put_pixel5(i, i, 0x7fff);
  }

  // Test Bresenham line algo.
  colour5 = 0x3c1f;
  draw_line5(0, 0, 127, 127);

  // Ooh and Midpoint circle algo :)
  colour5 = 0x7fe0;
  circle_wire5(32, 40, 30);

  // Umm maybe ellipses look better aspect ratio...
  ellipse_wire5(96, 40, 16, 30);
  ellipse_wire5(96, 120, (30 >> 1) + (30 >> 4), 30);
  circle_wire5(32, 120, 30);

  colour5 = 0x7fff;
  circle_wire5(64, 80, 64);
  //ellipse5(64, 80, 34, 64);
  
  // Test circle is right ratio...
  for(x = 54; x < 74; x ++) {
    put_pixel5(x, 80 - 64, 0x1f);
    put_pixel5(x, 80 + 64, 0x1f);
  }
  for(y = 70; y < 90; y ++) {
    put_pixel5(64 - 34, y, 0x1f);
    put_pixel5(64 + 34, y, 0x1f);
  }

  for(i = 60; i > 8; i -= 4) {
    colour5 = gba_rand();
    circle_fill5(64, 80, i);
  }

  //colour5 = 0x03e0;
  //ellipse_wire5(64, 80, (64 >> 1) + (64 >> 4), 64);

  // Try flicking scale on/off
  while(0) {
    BG2PC = 256;
    for(i = 0; i < 10; i ++) gba_vsync();
    BG2PC = 137;
    for(i = 0; i < 10; i ++) gba_vsync();
  }

  // Sphere lookylike?
  /*for(i = 0; i < 32; i ++) {
    colour5 = (31 - i) << 5;
    circle_fill5(64 + (i >> 1), 80 + (i >> 1), 64 - 2 * i);
    //ellipse5(64, 80, (64 >> 1) + (64 >> 4), 64);
  }*/

  while(1);

  return(0);
}
