#include "gba.h"
#include "stdlib.h"

// Declare any helper functions.
//void circle_plot5(u32 xc, u32 yc, u32 x, u32 y);
void ellipse_fill(u32 xc, u32 yc, u32 x, u32 y);
void ellipse_wire(u32 xc, u32 yc, u32 x, u32 y);
void scanline5(u32 x, u32 y, u32 len);

u16 *screen_bank5 = (u16 *)0x06000000;
u32 colour5 = 0x7fff7fff;

void clear5(u32 col) {
  u32 *screen = (u32 *)screen_bank5;
  int i;

  for(i = 0; i < 1024; i ++) {
    *screen++ = col;
    *screen++ = col;
    *screen++ = col;
    *screen++ = col;
    *screen++ = col;
    *screen++ = col;
    *screen++ = col;
    *screen++ = col;
    *screen++ = col;
    *screen++ = col;
  }
}

// Place specified color pixel at x,y (0 < x < 128, 0 < y < 160).
void put_pixel5(u32 x, u32 y, u32 col) {
  screen_bank5[x * 160 + y] = (u16)col;
}

// Place current color pixel at x,y (0 < x < 128, 0 < y < 160).
void draw_pixel5(u32 x, u32 y) {
  screen_bank5[x * 160 + y] = (u16)colour5;
}

// Draw line between and including two specified points.
// This implementation uses Bresenham, but it can be faster to do simple
// incremental if you use a fast division lookup.
void draw_line5(u32 x1, u32 y1, u32 x2, u32 y2) {
  int dx = abs(x1 - x2);
  int dy = abs(y1 - y2);
  int p = 2 * dy - dx;
  int dy2 = 2 * dy;
  int dydx2 = 2 * (dy - dx);
  int x, y, x_end;

  if(x1 > x2) {
    x = x2;
    y = y2;
    x_end = x1;
  } else {
    x = x1;
    y = y1;
    x_end = x2;
  }

  screen_bank5[x * 160 + y] = colour5;
  
  while(x < x_end) {
    x ++;
    if(p < 0) {
      p += dy2;
    } else {
      y ++;
      p += dydx2;
    }
    screen_bank5[x * 160 + y] = colour5;
  }
}



// Draw circle centred at x, y with radius r pixels.
// midpoint algorithm.
/* NOT USEFUL DUE TO ASPECTRATIO ISSUES.
void circle5(u32 xc, u32 yc, u32 radius) {
  int x = 0;
  int y = radius;
  int p = 1 - radius;

  // Plot first set of points.
  circle_plot5(xc, yc, x, y);

  while(x < y) {
    x ++;
    if(p < 0) {
      p += 2 * x + 1;
    } else {
      y --;
      p += 2 * (x - y) + 1;
    }
    circle_plot5(xc, yc, x, y);
  }
}
*/



// Helper function for circle plotting.
/* NOT HELPFUL - SEE ABOVE.
void circle_plot5(u32 xc, u32 yc, u32 x, u32 y) {
  screen_bank5[(xc + x) * 160 + yc + y] = colour5;
  screen_bank5[(xc - x) * 160 + yc + y] = colour5;
  screen_bank5[(xc + x) * 160 + yc - y] = colour5;
  screen_bank5[(xc - x) * 160 + yc - y] = colour5;
  screen_bank5[(xc + y) * 160 + yc + x] = colour5;
  screen_bank5[(xc - y) * 160 + yc + x] = colour5;
  screen_bank5[(xc + y) * 160 + yc - x] = colour5;
  screen_bank5[(xc - y) * 160 + yc - x] = colour5;
}
*/

// Draw filled elipse centred at x, y with radii rx and rx pixels.
// midpoint algorithm.
void ellipse_fill5(u32 xc, u32 yc, u32 rx, u32 ry) {
  int rx_sq = rx * rx;
  int ry_sq = ry * ry;
  int two_rx_sq = 2 * rx_sq;
  int two_ry_sq = 2 * ry_sq;
  int p;
  int x = 0;
  int y = ry;
  int px = 0;
  int py = two_rx_sq * y;

  // Plot first set of points.
  ellipse_fill(xc, yc, x, y);

  // Region 1.
  p = ry_sq - (rx_sq * ry) + (rx_sq >> 2);	// NB Float fudge here.
  while(px < py) {
    x ++;
    px += two_ry_sq;
    if(p < 0) {
      p += ry_sq + px;
    } else {
      y --;
      py -= two_rx_sq;
      p += ry_sq + px - py;
    }
    ellipse_fill(xc, yc, x, y);
  }

  // Region 2.
  // NB Floating point fudge here.
  p = ry_sq * ((x * x) + x) + rx_sq * (y - 1) * (y - 1) - rx_sq * ry_sq;
  while(y > 0) {
    y --;
    py -= two_rx_sq;
    if(p > 0) {
      p += rx_sq - py;
    } else {
      x ++;
      px += two_ry_sq;
      p += rx_sq - py + px;
    }
    ellipse_fill(xc, yc, x, y);
  }
}

// Draw wireframe ellipse centred at x, y with radii rx and rx pixels.
// midpoint algorithm.
void ellipse_wire5(u32 xc, u32 yc, u32 rx, u32 ry) {
  int rx_sq = rx * rx;
  int ry_sq = ry * ry;
  int two_rx_sq = 2 * rx_sq;
  int two_ry_sq = 2 * ry_sq;
  int p;
  int x = 0;
  int y = ry;
  int px = 0;
  int py = two_rx_sq * y;

  // Plot first set of points.
  ellipse_wire(xc, yc, x, y);

  // Region 1.
  p = ry_sq - (rx_sq * ry) + (rx_sq >> 2);	// NB Float fudge here.
  while(px < py) {
    x ++;
    px += two_ry_sq;
    if(p < 0) {
      p += ry_sq + px;
    } else {
      y --;
      py -= two_rx_sq;
      p += ry_sq + px - py;
    }
    ellipse_wire(xc, yc, x, y);
  }

  // Region 2.
  // NB Floating point fudge here.
  p = ry_sq * ((x * x) + x) + rx_sq * (y - 1) * (y - 1) - rx_sq * ry_sq;
  while(y > 0) {
    y --;
    py -= two_rx_sq;
    if(p > 0) {
      p += rx_sq - py;
    } else {
      x ++;
      px += two_ry_sq;
      p += rx_sq - py + px;
    }
    ellipse_wire(xc, yc, x, y);
  }
}

// Helper function for ellipse wireframe plotting.
void ellipse_wire(u32 xc, u32 yc, u32 x, u32 y) {
  screen_bank5[(xc + x) * 160 + yc + y] = colour5;
  screen_bank5[(xc - x) * 160 + yc + y] = colour5;
  screen_bank5[(xc + x) * 160 + yc - y] = colour5;
  screen_bank5[(xc - x) * 160 + yc - y] = colour5;
}

// Helper function for ellipse filled plotting.
void ellipse_fill(u32 xc, u32 yc, u32 x, u32 y) {
  //draw_line5(xc + x, yc + y, xc - x, yc + y);
  //draw_line5(xc + x, yc - y, xc - x, yc - y);
  scanline5(xc - x, yc + y, x * 2);
  scanline5(xc - x, yc - y, x * 2);
}

// Attempt at a fast scanline function.
void scanline5(u32 x, u32 y, u32 len) {
  u16 *start = screen_bank5 + x * 160 + y;

  switch(len) {
    case 128: *start = colour5; start += 160;
    case 127: *start = colour5; start += 160;
    case 126: *start = colour5; start += 160;
    case 125: *start = colour5; start += 160;
    case 124: *start = colour5; start += 160;
    case 123: *start = colour5; start += 160;
    case 122: *start = colour5; start += 160;
    case 121: *start = colour5; start += 160;
    case 120: *start = colour5; start += 160;
    case 119: *start = colour5; start += 160;
    case 118: *start = colour5; start += 160;
    case 117: *start = colour5; start += 160;
    case 116: *start = colour5; start += 160;
    case 115: *start = colour5; start += 160;
    case 114: *start = colour5; start += 160;
    case 113: *start = colour5; start += 160;
    case 112: *start = colour5; start += 160;
    case 111: *start = colour5; start += 160;
    case 110: *start = colour5; start += 160;
    case 109: *start = colour5; start += 160;
    case 108: *start = colour5; start += 160;
    case 107: *start = colour5; start += 160;
    case 106: *start = colour5; start += 160;
    case 105: *start = colour5; start += 160;
    case 104: *start = colour5; start += 160;
    case 103: *start = colour5; start += 160;
    case 102: *start = colour5; start += 160;
    case 101: *start = colour5; start += 160;
    case 100: *start = colour5; start += 160;
    case 99: *start = colour5; start += 160;
    case 98: *start = colour5; start += 160;
    case 97: *start = colour5; start += 160;
    case 96: *start = colour5; start += 160;
    case 95: *start = colour5; start += 160;
    case 94: *start = colour5; start += 160;
    case 93: *start = colour5; start += 160;
    case 92: *start = colour5; start += 160;
    case 91: *start = colour5; start += 160;
    case 90: *start = colour5; start += 160;
    case 89: *start = colour5; start += 160;
    case 88: *start = colour5; start += 160;
    case 87: *start = colour5; start += 160;
    case 86: *start = colour5; start += 160;
    case 85: *start = colour5; start += 160;
    case 84: *start = colour5; start += 160;
    case 83: *start = colour5; start += 160;
    case 82: *start = colour5; start += 160;
    case 81: *start = colour5; start += 160;
    case 80: *start = colour5; start += 160;
    case 79: *start = colour5; start += 160;
    case 78: *start = colour5; start += 160;
    case 77: *start = colour5; start += 160;
    case 76: *start = colour5; start += 160;
    case 75: *start = colour5; start += 160;
    case 74: *start = colour5; start += 160;
    case 73: *start = colour5; start += 160;
    case 72: *start = colour5; start += 160;
    case 71: *start = colour5; start += 160;
    case 70: *start = colour5; start += 160;
    case 69: *start = colour5; start += 160;
    case 68: *start = colour5; start += 160;
    case 67: *start = colour5; start += 160;
    case 66: *start = colour5; start += 160;
    case 65: *start = colour5; start += 160;
    case 64: *start = colour5; start += 160;
    case 63: *start = colour5; start += 160;
    case 62: *start = colour5; start += 160;
    case 61: *start = colour5; start += 160;
    case 60: *start = colour5; start += 160;
    case 59: *start = colour5; start += 160;
    case 58: *start = colour5; start += 160;
    case 57: *start = colour5; start += 160;
    case 56: *start = colour5; start += 160;
    case 55: *start = colour5; start += 160;
    case 54: *start = colour5; start += 160;
    case 53: *start = colour5; start += 160;
    case 52: *start = colour5; start += 160;
    case 51: *start = colour5; start += 160;
    case 50: *start = colour5; start += 160;
    case 49: *start = colour5; start += 160;
    case 48: *start = colour5; start += 160;
    case 47: *start = colour5; start += 160;
    case 46: *start = colour5; start += 160;
    case 45: *start = colour5; start += 160;
    case 44: *start = colour5; start += 160;
    case 43: *start = colour5; start += 160;
    case 42: *start = colour5; start += 160;
    case 41: *start = colour5; start += 160;
    case 40: *start = colour5; start += 160;
    case 39: *start = colour5; start += 160;
    case 38: *start = colour5; start += 160;
    case 37: *start = colour5; start += 160;
    case 36: *start = colour5; start += 160;
    case 35: *start = colour5; start += 160;
    case 34: *start = colour5; start += 160;
    case 33: *start = colour5; start += 160;
    case 32: *start = colour5; start += 160;
    case 31: *start = colour5; start += 160;
    case 30: *start = colour5; start += 160;
    case 29: *start = colour5; start += 160;
    case 28: *start = colour5; start += 160;
    case 27: *start = colour5; start += 160;
    case 26: *start = colour5; start += 160;
    case 25: *start = colour5; start += 160;
    case 24: *start = colour5; start += 160;
    case 23: *start = colour5; start += 160;
    case 22: *start = colour5; start += 160;
    case 21: *start = colour5; start += 160;
    case 20: *start = colour5; start += 160;
    case 19: *start = colour5; start += 160;
    case 18: *start = colour5; start += 160;
    case 17: *start = colour5; start += 160;
    case 16: *start = colour5; start += 160;
    case 15: *start = colour5; start += 160;
    case 14: *start = colour5; start += 160;
    case 13: *start = colour5; start += 160;
    case 12: *start = colour5; start += 160;
    case 11: *start = colour5; start += 160;
    case 10: *start = colour5; start += 160;
    case 9: *start = colour5; start += 160;
    case 8: *start = colour5; start += 160;
    case 7: *start = colour5; start += 160;
    case 6: *start = colour5; start += 160;
    case 5: *start = colour5; start += 160;
    case 4: *start = colour5; start += 160;
    case 3: *start = colour5; start += 160;
    case 2: *start = colour5; start += 160;
    case 1: *start = colour5; start += 160;
    case 0: *start = colour5; start += 160;	// Even 0 length plots 1 pixel.
    default: break;
  }
}
